/*
 * Copyright (C) 1998-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "gui.cmdline.h"
#include "para.h"

#include <curses.h>
#include <sys/time.h> /* timeval, needed for select */


#define STANDARD_STATUS_BAR "para_gui " VERSION " (hit ? for help)"
#define MAX_WRITE_STREAM_DELAY 3
#define MIN_TOP_LINES 9

const char *logo="       / \n"
"      /    _ __   __ _ _ __ __ _ ___| | __ _ ___| |__\n"
"  ####    | '_ \\ / _` | '__/ _` / __| |/ _` / __| '_ \\\n"
"  #  #    | |_) | (_| | | | (_| \\__ \\ | (_| \\__ \\ | | |\n"
"  ####    | .__/ \\__,_|_|  \\__,_|___/_|\\__,_|___/_| |_|\n"
" /        |_|\n"
"/";

static void finish(int sig);

static struct win_data {
	WINDOW *win;
	NCURSES_SIZE_T begx;
	NCURSES_SIZE_T begy;
	NCURSES_SIZE_T cols;
	NCURSES_SIZE_T lines;
} top, bot, sb, in, sep;

static int got_usr1, got_deadly_signal, got_winch, got_int, got_chld,
	external_cmd_died, curses_active, current_audio_format = -1;
static pid_t external_cmd_pid;

static FILE *command_pipe, *stat_pipe;


static struct gengetopt_args_info args_info;

enum {GETCH_MODE, COMMAND_MODE, EXTERNAL_MODE};

#define LEFT 1
#define RIGHT 2
#define CENTER 3


#define COLOR_STATUSBAR 32
#define COLOR_COMMAND 33
#define COLOR_OUTPUT 34
#define COLOR_MSG 35
#define COLOR_ERRMSG 36
#define COLOR_WELCOME 37
#define COLOR_SEPARATOR 38

#define USECS 1000000
#define MSECS 1000

struct stat_item{
	char name[MAXLINE];
	char prefix[MAXLINE];
	char postfix[MAXLINE];
	unsigned y;
	unsigned x;
	unsigned len;
	int fg, bg;
	int align;
	char content[MAXLINE];
};

struct stream_io{
	char *name;
	char *rcmd, *wcmd;
	pid_t rpid, wpid;
	FILE *read, *write;
	int rdied, wdied;
	char *buf;
	int bufsize, buflen;
	struct timeval rtime, wtime;
};

static struct stream_io siol[] = { /* stream io list */
#if 1
	{
		.name = "mp3",
		.rcmd = "pob-fec -q -b 2",
		.wcmd = "madplay -",
		.buf = NULL,
		.bufsize = 4096,
		.read = NULL,
		.write = NULL,
		.rdied = 0,
		.wdied = 0
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.rcmd = "para_pob-ogg",
		.wcmd = "ogg123 -",
		.buf = NULL,
		.bufsize = 4096,
		.read = NULL,
		.write = NULL,
		.rdied = 0,
		.wdied = 0
	},
#endif
	{
		.name = NULL,
	}
};

enum {STATUS_BAR, STATUS, STATUS_FLAGS, NUM_PLAYED, STREAM, LENGTH_MIN, LENGTH_SEC,
	SCORE, MTIME, FILE_SIZE, FORMAT,
	AUDIO_INFO1, AUDIO_INFO2,AUDIO_INFO3, 
	DBINFO1, DBINFO2, DBINFO3};

struct stat_item stat_items[] = {
	{
		.name = "status_bar",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_CYAN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 7,
		.len = 100,
	},
	{
		.name = "status",
		.prefix = "status: ",
		.postfix = " ",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = RIGHT,
		.content = "",
		.x = 0,
		.y = 17,
		.len = 35,
	},
	{
		.name = "status_flags",
		.prefix = "(",
		.postfix = ")",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = LEFT,
		.content = "",
		.x = 35,
		.y = 17,
		.len = 10,
	},
	{
		.name = "num_played",
		.prefix = "#",
		.postfix = "",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 45,
		.y = 17,
		.len = 10,
	},
	{
		.name = "stream",
		.prefix = "stream: ",
		.postfix = "",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = LEFT,
		.content = "",
		.x = 72,
		.y = 17,
		.len = 28,
	},
	{
		.name = "length_min",
		.prefix = "Length: ",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = RIGHT,
		.content = "",
		.x = 0,
		.y = 27,
		.len = 15,
	},
	{
		.name = "length_sec",
		.prefix = ":",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = LEFT,
		.content = "",
		.x = 15,
		.y = 27,
		.len = 8,
	},
	{
		.name = "score",
		.prefix = "Score: ",
		.postfix = "",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 55,
		.y = 17,
		.len = 17,
	},
	{
		.name = "mtime",
		.prefix = "Mtime: ",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 23,
		.y = 27,
		.len = 40,
	},
	{
		.name = "file_size",
		.prefix = "Size: ",
		.postfix = " kb",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 63,
		.y = 27,
		.len = 20,
	},
	{
		.name = "format",
		.prefix = "format: ",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 84,
		.y = 27,
		.len = 16,
	},
	{
		.name = "audio_file_info1",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_GREEN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 43,
		.len = 100,
	},
	{
		.name = "audio_file_info2",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_GREEN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 53,
		.len = 100,
	},
	{
		.name = "audio_file_info3",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_GREEN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 63,
		.len = 100,
	},
	{
		.name = "dbinfo1",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_YELLOW,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 77,
		.len = 100,
	},
	{
		.name = "dbinfo2",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_YELLOW,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 87,
		.len = 100,
	},
	{
		.name = "dbinfo3",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_YELLOW,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 97,
		.len = 100,
	},
	{
		.name = "",
	}
};

int _argc;
char **_argv;

static void com_help(int);
static void com_reread_conf(int);
static void com_sl(int);
static void com_toggle_auto_decode(int);
static void com_enlarge_top_win(int);
static void com_shrink_top_win(int);
static void com_version(int);
static void com_query_stream(int);
static void com_quit(int);
static void com_refresh(int);
static void com_ll_incr(int);
static void com_ll_decr(int);

struct gui_command command_list[] = {
	{
		.key = "?",
		.name = "help",
		.description = "print help",
		.handler = com_help
	}, {
		.key = "+",
		.name = "enlarge_win",
		.description = "enlarge the top window",
		.handler = com_enlarge_top_win
	}, {
		.key = "-",
		.name = "shrink_win",
		.description = "shrink the top window",
		.handler = com_shrink_top_win
	}, {
		.key = "d",
		.name = "auto_decode",
		.description = "toggle auto decoding",
		.handler = com_toggle_auto_decode
	}, {
		.key = "r",
		.name = "reread_conf",
		.description = "reread configuration file",
		.handler = com_reread_conf
	}, {
		.key = "f",
		.name = "sl",
		.description = "print next 10 songs",
		.handler = com_sl
	}, {
		.key = "q",
		.name = "quit",
		.description = "kill decoder and exit",
		.handler = com_quit
	}, {
		.key = "^L",
		.name = "refresh",
		.description = "redraw the screen",
		.handler = com_refresh
	}, {
		.key = "Q",
		.name = "query_stream",
		.description = "print definition of current stream",
		.handler = com_query_stream
	}, {
		.key = ">",
		.name = "ll_incr",
		.description = "increase loglevel (decreases verbosity)",
		.handler = com_ll_incr
	}, {
		.key = "<",
		.name = "ll_decr",
		.description = "decrease loglevel (increases verbosity)",
		.handler = com_ll_decr
	}, {
		.key = "V",
		.name = "version",
		.description = "show the para_gui version",
		.handler = com_version
	}, {
		.handler = NULL
	}
};

static int find_cmd_byname(char *name)
{
	int i;

	for (i = 0; command_list[i].handler; i++)
		if (!strcmp(command_list[i].name, name))
			return i;
	return -1;
}

/* taken from mutt */
static char *km_keyname (int c)
{
	static char buf[10];

	if (c < 256 && c > -128 && iscntrl ((unsigned char) c)) {
		if (c < 0)
			c += 256;
		if (c < 128) {
			buf[0] = '^';
			buf[1] = (c + '@') & 0x7f;
			buf[2] = 0;
		} else
			snprintf (buf, sizeof (buf), "\\%d%d%d", c >> 6,
				(c >> 3) & 7, c & 7);
	} else if (c >= KEY_F0 && c < KEY_F(256))
		sprintf (buf, "<F%d>", c - KEY_F0);
	else if (isprint (c))
		snprintf (buf, sizeof (buf), "%c", (unsigned char) c);
	else
		snprintf (buf, sizeof (buf), "\\x%hx", (unsigned short) c);
	return (buf);
}

static char *configfile_exists(void)
{
	static char *config_file;
	char *tmp;

	if (!args_info.config_file_given) {
		if (!config_file)
			config_file = malloc(_POSIX_PATH_MAX * sizeof(char));
		sprintf(config_file, "%s/.paraslash/gui.conf", getenv("HOME"));
		tmp = config_file;
	} else
		tmp = args_info.config_file_arg;
	return file_exists(tmp)? tmp: NULL;
}

/*
 * print num spaces to curses window
 */
static void add_spaces(WINDOW* win, unsigned int num)
{
	while (num > 0) {
		num--;
		waddstr(win, " ");
	}
}

/*
 * print aligned string to curses window. This function always prints
 * exactly len chars.
 */
static int align_str(WINDOW* win, char *string, unsigned int len,
		unsigned int align)
{
	int num;
	char *str;

	if (!win || !string)
		return -1;
	if (!(str = strdup(string)))
		return -1;
	num = len - strlen(str);	
	if (num < 0) {
		str[len] = '\0';
		num = 0;
	}
	if (align == LEFT) {
		waddstr(win, str);
		add_spaces(win, num);
	} else if (align == RIGHT) {
		add_spaces(win, num);
		waddstr(win, str);
	} else {
		add_spaces(win, num / 2);
		waddstr(win, str[0]? str: "");
		add_spaces(win, num - num / 2);
	}
	free(str);
	return 1;
}

static void print_in_bar(int color, char *fmt,...)
{

	va_list argp;
	static char buf[MAXLINE];

	if (external_cmd_pid)
		return; /* curses not active */
	wattron(in.win, COLOR_PAIR(color));
	va_start(argp, fmt);
	vsprintf(buf, fmt, argp);
	va_end(argp);
	wmove(in.win, 0, 0);
	align_str(in.win, buf? buf: " ", sb.cols, LEFT);
	wrefresh(in.win);
}

/*
 * update the status bar
 */
static void print_status_bar(char *line)
{
	wmove(sb.win, 0, 0);
	align_str(sb.win, line? line: STANDARD_STATUS_BAR, sb.cols, CENTER);
	wrefresh(sb.win);
}

static void add_output_line(char *line)
{
	waddstr(bot.win, line);
	waddstr(bot.win, "\n");
	wrefresh(bot.win);
}

static void deadly_signal_handler(int i)
{
	/* this will kill us in do_select */
	got_deadly_signal = i;
}

static void sigwinch_handler(int i)
{
	got_winch = 1;
}

static void sigint_handler(int i)
{
	got_int = 1;
}

static void sigusr1_handler(int i)
{
	got_usr1 = 1;
}

static void sigchld_handler(int irrelevant)
{
	pid_t pid;
	int status, i;

	got_chld = 1;
	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		if (pid == external_cmd_pid) {
			external_cmd_pid = 0;
			external_cmd_died = 1;
			continue;
		}
		for (i = 0; siol[i].name; i++) {
			if (pid == siol[i].rpid)
				siol[i].rdied = 1;
			else if (pid == siol[i].wpid)
				siol[i].wdied = 1;
		}
	}
}

static void signals_setup(void)
{
	signal(SIGINT, sigint_handler);
	signal(SIGTERM, deadly_signal_handler);
	signal(SIGCHLD, sigchld_handler);
	signal(SIGWINCH, sigwinch_handler);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGUSR1, sigusr1_handler);
	signal(SIGHUP, deadly_signal_handler);
}

static void kill_children(void)
{
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	signal(SIGTERM, deadly_signal_handler);
}

static void do_exit(int ret)
{
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	exit(ret);
}

static void shutdown_curses(void)
{
	def_prog_mode();
	curses_active = 0;
	endwin();
}

static void finish(int ret)
{
	shutdown_curses();
	do_exit(ret);
}

/*
 * exit curses and print given message
 */
static void msg_n_exit(int ret, char* fmt,...)
{
	va_list argp;
	FILE *outfd = ret? stderr: stdout;

	shutdown_curses();
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
	do_exit(ret);
}

/*
 * replace first '\n' by '\0' and return pointer to char after first ':'
 * in str.
 */
static char *prune_str(char *str, unsigned int len)
{
	char *p = str, *p1;
	unsigned int i = 0;

	while (*p != ':') {
		p++;
		if (*p == '\0')
			return NULL;
	}
	p++;
	p1 = p;
	while (*p1 != '\0' && i < len) {
		if (*p1 == '\n')
			break;
		p1++;
		i++;
	}
	*p1 = '\0';
	return p;
}

/*
 * print formated output to bot win
 */
static void outputf(int color, char* fmt,...)
{
	va_list argp;
	static char buf[MAXLINE];

	if (external_cmd_pid)
		return; /* curses not active */
	wattron(bot.win, COLOR_PAIR(color));
	va_start(argp, fmt);
	vsprintf(buf, fmt, argp);
	va_end(argp);
	waddstr(bot.win, buf);
	wrefresh(bot.win);
}

static void clog(int ll, char *fmt,...)
{
	int color;
	va_list argp;
	static char buf[MAXLINE];

	if (ll < args_info.loglevel_arg)
		return;
	switch (ll) {
		case DEBUG:
		case VERBOSE:
		case INFO:
			color = COLOR_MSG;
			break;
		default:
			color = COLOR_ERRMSG;
	}
	va_start(argp, fmt);
	vsprintf(buf, fmt, argp);
	va_end(argp);
	if (curses_active) {
		wattron(bot.win, COLOR_PAIR(color));
		waddstr(bot.win, buf);
		wrefresh(bot.win);
	} else 
		fprintf(stderr, "%s", buf);
}

static void print_welcome(void)
{
	outputf(COLOR_WELCOME, "Welcome to para_gui " VERSION
		" \"" CODENAME "\"\n");
	wclrtoeol(bot.win);
	if (bot.lines > 23)
		outputf(COLOR_MSG, "%s\n", logo);
}

/*
 * int all windows
 */
static void init_wins(int top_lines)
{
	int i;

	top.lines = top_lines;
	top.cols = COLS;
	top.begy = 0;
	top.begx = 0;

	bot.lines = LINES - top.lines - 3;
	bot.cols = COLS;
	bot.begy = top.lines + 1;
	bot.begx = 0;

	sb.lines = 1;
	sb.cols = COLS;
	sb.begy = LINES - 2;
	sb.begx = 0;
	
	in.lines = 1;
	in.cols = COLS;
	in.begy = LINES - 1;
	in.begx = 0;

	sep.lines = 1;
	sep.cols = COLS;
	sep.begy = top.lines;
	sep.begx = 0;
	
	if (top.win) {
		mvwin(top.win, top.begy, top.begx);
		wresize(top.win, top.lines, top.cols);

		mvwin(sb.win, sb.begy, sb.begx);
		wresize(sb.win, sb.lines, sb.cols);

		mvwin(sep.win, sep.begy, sep.begx);
		wresize(sep.win, sep.lines, sep.cols);

		mvwin(bot.win, bot.begy, bot.begx);
		wresize(bot.win, bot.lines, bot.cols);

		mvwin(in.win, in.begy, in.begx);
		wresize(in.win, in.lines, in.cols);
	} else {
		sep.win = newwin(sep.lines, sep.cols, sep.begy, sep.begx);
		top.win = newwin(top.lines, top.cols, top.begy, top.begx);
		bot.win = newwin(bot.lines, bot.cols, bot.begy, bot.begx);
		sb.win = newwin(sb.lines, sb.cols, sb.begy, sb.begx);
		in.win = newwin(in.lines, in.cols, in.begy, in.begx);
		if (!top.win || !bot.win || !sb.win || !in.win || !sep.win)
			msg_n_exit(1, "Error: Cannot create curses windows\n");
		wclear(bot.win);
		wclear(top.win);
		wclear(sb.win);
		wclear(in.win);
		scrollok(bot.win, 1);
		wattron(sb.win, COLOR_PAIR(COLOR_STATUSBAR));
		wattron(sep.win, COLOR_PAIR(COLOR_SEPARATOR));
		nodelay(top.win, 1);
		nodelay(bot.win, 1);
		nodelay(sb.win, 1);
		nodelay(in.win, 0);

		keypad(top.win, 1);
		keypad(bot.win, 1);
		keypad(sb.win, 1);
		keypad(in.win, 1);
		print_status_bar(NULL);
	}
	
	for (i = 1; i <= COLS; i++)
		 waddstr(sep.win, "-");

	wclear(top.win);
	wnoutrefresh(top.win);
	wnoutrefresh(bot.win);
	//wnoutrefresh(sb.win);
	print_status_bar(NULL);
	wnoutrefresh(in.win);
	wnoutrefresh(sep.win);
	doupdate();
}

static int top_lines_ok(int top_lines)
{
	return (top_lines < MIN_TOP_LINES || LINES - top_lines < 5)? 0:1;
}

static void check_geometry(void)
{
	if (LINES < 13 || COLS < 60)
		msg_n_exit(EXIT_FAILURE, "Error: Terminal (%dx%d) too narrow"
			" (need at least 60x13)\n", COLS, LINES);
}

/*
 * Print item #i to curses window
 */
static void print_item(int i)
{
	struct stat_item *s = stat_items + i;
	char buf[MAXLINE];

	if (!curses_active)
		return;
	if (!s->content) {
		clog(ERROR, "s->content == NULL\n");
		return;
	}
	snprintf(buf, MAXLINE - 1, "%s%s%s", s->prefix, s->content, s->postfix);
	buf[MAXLINE] = '\0';
	wmove(top.win, s->y * top.lines / 100, s->x * COLS / 100);
	wrefresh(top.win);
	wattron(top.win, COLOR_PAIR(i + 1));
	align_str(top.win, buf, s->len * COLS / 100, s->align);
	wrefresh(top.win);
}

static void print_all_items(void)
{
	int i;

	for (i = 0; stat_items[i].name[0]; i++)
		print_item(i);
}

static void init_colors(void)
{
	int i;
	if (!has_colors())
		msg_n_exit(EXIT_FAILURE, "Error: No color term\n");
	start_color();

	for (i = 0; stat_items[i].name[0]; i++)
		init_pair(i + 1, stat_items[i].fg, stat_items[i].bg);

	init_pair(COLOR_STATUSBAR, COLOR_BLACK, COLOR_GREEN);
	init_pair(COLOR_COMMAND, COLOR_YELLOW, COLOR_BLACK);
	init_pair(COLOR_OUTPUT, COLOR_CYAN, COLOR_BLACK);
	init_pair(COLOR_MSG, COLOR_WHITE, COLOR_BLACK);
	init_pair(COLOR_ERRMSG, COLOR_WHITE, COLOR_RED);
	init_pair(COLOR_WELCOME, COLOR_WHITE, COLOR_BLUE);
	init_pair(COLOR_SEPARATOR, COLOR_BLUE, COLOR_BLACK);
}

/*
 * (re-)initialize the curses library
 */
static void init_curses(void)
{
	//resize_term(LINES, COLS);
	refresh(); /* really needed */
	check_geometry();
	curs_set(0); /* really needed */
	noraw();
	nonl(); /* tell curses not to do NL->CR/NL on output */
	noecho(); /* don't echo input */
	cbreak(); /* take input chars one at a time, no wait for \n */
	//reset_prog_mode();
	init_colors();
	curses_active = 1;
	init_wins(top_lines_ok(top.lines)? top.lines : MIN_TOP_LINES);
	print_all_items();
	noecho(); /* don't echo input */
}

static void start_stream_writer(int i)
{ 
	struct stream_io *s = &siol[i];

	clog(DEBUG, "starting %s stream writer (%s)\n", s->name, s->wcmd);
	s->write = popen_write_cmdline_pid(&s->wpid, s->wcmd);
	/* we write to this fd in do_select, so we need non-blocking */
	fcntl(fileno(s->write), F_SETFL, O_NONBLOCK);
	gettimeofday(&s->wtime, NULL);
	clog(DEBUG, "%s stream_writer started, pid = %d\n", s->name, s->wpid);
}

static void start_stream_reader(int i)
{
	struct stream_io *s = &siol[i];

	clog(DEBUG, "starting %s stream reader (%s)\n", s->name, s->rcmd);
	s->read = popen_read_cmdline_pid(&s->rpid, s->rcmd);
	s->buflen = 0;
	gettimeofday(&s->rtime, NULL);
	clog(DEBUG, "%s stream reader started, pid = %d\n", s->name, s->rpid);
}

static void kill_stream_writer(int i)
{
	struct stream_io *s = &siol[i];

	if (s->write) {
		clog(DEBUG, "stopping %s stream writer\n", s->name);
		kill(s->wpid, SIGTERM);
		s->buflen = 0;
	}
}

static void kill_stream_reader(int i)
{
	struct stream_io *s = &siol[i];

	if (s->read) {
		clog(INFO, "stopping %s stream reader\n", s->name);
		kill(s->rpid, SIGTERM);
		s->buflen = 0;
	}
}
/*
 * Send SIGTERM to reader and writer if present
 */
static void kill_decoder(int i)
{
	struct stream_io *s = &siol[i];
	if (!s->write && !s->read)
		return;
	kill_stream_writer(i);
	kill_stream_reader(i);
	usleep(100000);
}

static void kill_all_decoders(void)
{
	int i;

	for (i = 0; siol[i].name; i++) 
		kill_decoder(i);
}

static void start_decoder(void)
{
	int i;
	struct stream_io *s;

	for (i = 0; siol[i].name; i++)
		if (!strcmp(stat_items[FORMAT].content, siol[i].name)) 
			break;
	if (!siol[i].name) {
		clog(ERROR, "unknown audio format\n");
		return;
	}
	s = &siol[i];
//	if (i != current_audio_format && current_audio_format >= 0) {
//		kill_decoder(current_audio_format);
//	}
	current_audio_format = i;
	
	if (!s->write)
		start_stream_writer(i);
	if (!s->read)
		start_stream_reader(i);
}

/*
 * extract contents from line which is already known to start with
 * name of i-th command to stat_items' content.
 */
static void print_stat_line(char *line, int i)
{
	char *strptr;
	struct stat_item *s = stat_items + i;
	
	if (!line) {
		clog(ERROR, "%s: line == NULL\n", __func__);
		return;
	}
	strptr = prune_str(line, s->len - strlen(s->prefix)
		- strlen(s->postfix));
	if (!strptr)
		return;
	strcpy(s->content, strptr);
	print_item(i);
}

/*
 * print status line if line starts with known command.
 */
static void check_stat_line(char *line)
{
	int i;
	size_t line_len = strlen(line);

	for (i = 0; stat_items[i].name[0]; i++) {
		struct stat_item *s = stat_items + i;
		size_t item_len = strlen(s->name);
		
		if (line_len > item_len && line[item_len] == ':' &&
 				!strncmp(line, s->name, item_len))
			print_stat_line(line, i);
	}
}


/*
 * This sucker modifies its first argument. *handler and *arg are
 * pointers to 0-terminated strings (inside line). Crap.
 */
static int split_key_map(char *line, char **handler, char **arg)
{
	if (!(*handler = strchr(line + 1, ':')))
		goto err_out;
	**handler = '\0';
	(*handler)++;
	if (!(*arg = strchr(*handler, ':')))
		goto err_out;
	**arg = '\0';
	(*arg)++;
	return 1;
err_out:
	return 0;
}

static int check_key_map_args(void)
{
	char *s;
	int i, ret = -1;
	char *tmp = NULL, *handler, *arg;

	for (i = 0; i < args_info.key_map_given; ++i) {
		s = args_info.key_map_arg[i];
		if (!(*s))
			goto err_out;
		if (tmp)
			free(tmp);
		tmp = strdup(s);
		if (!split_key_map(tmp, &handler, &arg))
			goto err_out;
		if (strlen(handler) != 1)
			goto err_out;
		if (*handler != 'x'
			&& *handler != 'd'
			&& *handler != 'i'
			&& *handler != 'p')
			goto err_out;
		if (*handler != 'i')
			continue;
		if (find_cmd_byname(arg) < 0)
			goto err_out;
	}
	ret = 0;
err_out:
	if (tmp)
		free(tmp);
	return ret;
}

/*
 * React to various signal-related events
 */
static void check_signals(void)
{
	int i;

	if (got_deadly_signal)
		msg_n_exit(EXIT_FAILURE,
			"Only the good die young (caught "
			"deadly signal)\n");
	if (got_winch) {
		got_winch = 0;
		if (!external_cmd_pid) {
			shutdown_curses();
			init_curses();
		}
	}
	if (got_int) {
		clog(WARNING, "Caught SIGINT, killing children\n");
		got_int = 0;
		kill_children();
	}
	if (got_usr1) {
		got_usr1 = 0;
		clog(NOTICE, "Got SIGUSR1, rereading configuration\n");
		com_reread_conf(0);
	}
	for (i = 0; siol[i].name; i++) {
		struct stream_io *s = &siol[i];
		if (s->rdied) {
			s->rdied = 0;
			clog(INFO, "%s stream reader died\n", s->name);
			pclose(s->read);
			s->read = NULL;
			kill_stream_writer(i);
			usleep(300000);
		}
	}
	for (i = 0; siol[i].name; i++) {
		struct stream_io *s = &siol[i];
		if (s->wdied) {
			s->wdied = 0;
			clog(INFO, "%s stream writer died\n", s->name);
			pclose(s->write);
			s->write = NULL;
			kill_stream_reader(i);
			usleep(300000);
		}
	}
}

static void open_stat_pipe_msg(void)
{
	stat_pipe = open_stat_pipe();
	clog(INFO, "status pipe opened\n");
}

static int set_stream_fds(fd_set *rfds, fd_set *wfds)
{
	int i, max_fileno = 0;
	if (!args_info.auto_decode_flag)
		return 0;
	for (i = 0; siol[i].name; i++) {
		struct stream_io *s = &siol[i];
		int file_no;

		if (!s->write || !s->read)
			continue;
		if (s->buflen) {
			/* 
                         * buf already contains data which was not yet
                         * witten
			 */
			file_no = fileno(s->write);
			FD_SET(file_no, wfds);
			max_fileno = MAX(file_no, max_fileno);
		} else {
			file_no = fileno(s->read);
			FD_SET(file_no, rfds);
			max_fileno = MAX(file_no, max_fileno);
		}
	}
	return max_fileno;
}

static int write_audio_data(int format)
{
	struct stream_io *s = &siol[format];
	int rv;

	//clog(DEBUG, "writing %d byte to %s pipe\n", s->buflen, s->name);
	rv = write(fileno(s->write), s->buf, s->buflen);
	if (rv < 0) {
		clog(errno == EPIPE? INFO : ERROR, "write error: %s\n", 
			strerror(errno));
		kill_decoder(format);
		s->buflen = 0;	
	} else if (rv < s->buflen) {
		clog(DEBUG, "partial %s write (%i/%i)\n", s->name, rv, 
			s->buflen);
		s->buflen -= rv;
		memmove(s->buf, s->buf + rv, s->buflen);
	} else {
		gettimeofday(&s->wtime, NULL);
		s->buflen = 0;	
	}
	return rv;
}

static int read_audio_data(int format)
{
	struct stream_io *s = &siol[format];
	int file_no = fileno(s->read);

	//clog(DEBUG, "reading at most %d byte from %s pipe (fileno %d) to %lx\n", 
	//		s->bufsize, s->name, file_no, s->buf);
	s->buflen = read(file_no, s->buf, s->bufsize);
	if (s->buflen < 0) {
		clog(ERROR, "%s read error (%s)\n", s->name, strerror(errno));
		kill_decoder(format);
		s->buflen = 0;
	} else if (s->buflen > 0)
		gettimeofday(&s->rtime, NULL);
	return s->buflen;
}

static void check_read_time(int format)
{
	struct timeval now;
	struct stream_io *s = &siol[format];

	gettimeofday(&now, NULL);
	if (now.tv_sec <= s->rtime.tv_sec + MAX_WRITE_STREAM_DELAY)
		return;
	clog(INFO, "%s input buffer not ready, deactivating %s decoder\n",
			s->name, s->name);
	kill_decoder(format);
}
static void check_write_time(int format)
{
	struct timeval now;
	struct stream_io *s = &siol[format];

	gettimeofday(&now, NULL);
	if (now.tv_sec <= s->wtime.tv_sec + MAX_WRITE_STREAM_DELAY)
		return;
	clog(INFO, "%s output buffer not ready, deactivating %s decoder\n",
			s->name, s->name);
	kill_decoder(format);
}

static void do_stream_io(fd_set *rfds, fd_set *wfds)
{
	int i;
	for (i = 0; siol[i].name; i++) {
		struct stream_io *s = &siol[i];
		int file_no;
	
		if (!s->write || !s->read || !args_info.auto_decode_flag)
			continue;

		if (s->buflen) {
			file_no = fileno(s->write);
			if (FD_ISSET(file_no, wfds)) {
				write_audio_data(i);
				continue;
			}
			check_read_time(i);
		} else {
			file_no = fileno(s->read);
			if (FD_ISSET(file_no, rfds)) {
				read_audio_data(i);
				continue;
			}
			check_write_time(i);
		}
	}
}
static int stream_write_pending(void)
{
	int i;
	for (i = 0; siol[i].name; i++)
		if (siol[i].buflen > 0)
			return 1;
	return 0;
}
/*
 * This is the core function, dirty and mean. It has these modes:
 *
 * GETCH_MODE: check stdin and stat_pipe, return when key is
 * pressed
 *
 * COMMAND_MODE: check stat_pipe and command_pipe. Return when
 * a screen full of output has been read or when other end has closed command
 * pipe or when any key is pressed.
 *
 * EXTERNAL_MODE: Read/Write only stream pipe. Works with curses disabled.
 * Returns when external_cmd_pid == 0.
 */
static int do_select(int mode)
{
	fd_set rfds, wfds;
	/* stat pipe, command pipe, decoder */
	int sp_fileno, cp_fileno;
	int ret, lines = 0, total_lines = 0;
	int stream_buf_bytes = 0;
	int max_fileno, cp_numread = 1, sp_numread, playing;
	char status_buf[STRINGSIZE] = "";
	char command_buf[STRINGSIZE] = "";
	/* command/stat buf offset */
	int cbo = 0, sbo = 0;
repeat:
	playing = !strncmp(stat_items[STATUS].content, "playing", 7);
	if (!playing || !args_info.auto_decode_flag) {
		stream_buf_bytes = 0;
		//kill_decoder(current_audio_format);
	} else
		start_decoder();
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	max_fileno = set_stream_fds(&rfds, &wfds);
	/* stat pipe */
	if (stat_pipe) {
		sp_fileno = fileno(stat_pipe);
		FD_SET(sp_fileno, &rfds);
		max_fileno = MAX(max_fileno, sp_fileno);
	}
	/* command pipe only for COMMAND_MODE */
	if (mode == COMMAND_MODE && lines < bot.lines - 1) {
		cp_fileno = fileno(command_pipe);
		FD_SET(cp_fileno, &rfds);
		max_fileno = MAX(max_fileno, cp_fileno);
	}
	if (!max_fileno) {
		/*
		 * no fds to check, this means that an external command
		 * is running (otherwise we would check stat pipe) and
		 * we are not playing.
		 */
		sleep(1);
		ret = 0;
	} else {
		struct timeval tv;
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		ret = select(max_fileno + 1, &rfds, &wfds, NULL, &tv);
	}
	//clog(DEBUG, "select returned %d\n", ret);
	check_signals();
	if (ret <= 0)
		goto check_return; /* skip fd checks */
	if (!siol[current_audio_format].write)
		stream_buf_bytes = 0;
	do_stream_io(&rfds, &wfds);
	/* read status pipe */
	if (stat_pipe && FD_ISSET(sp_fileno, &rfds)) {
		if ((sp_numread = read(sp_fileno, status_buf + sbo,
				STRINGSIZE - 1 - sbo)) <= 0) {
			kill_all_decoders();
			clog(INFO, "closing status pipe\n");
			while (sleep(1) > 0)
				; /* try again*/
			pclose(stat_pipe);
			stat_pipe = NULL;
			open_stat_pipe_msg();
			while (sleep(1) > 0)
				; /* try again*/
		}
		sbo = check_buf_for_items(status_buf, sp_numread + sbo,
			&check_stat_line, 0);
	}
	/* read command pipe if ready */
	if (mode == COMMAND_MODE && lines < bot.lines - 1 &&
			FD_ISSET(cp_fileno, &rfds)) {
		cp_numread = read(cp_fileno, command_buf + cbo,
			STRINGSIZE - 1 - cbo);
		if (cp_numread >= 0)
			cbo += cp_numread;
		else if (cp_numread < 0)
			clog(ERROR, "read error\n");
	}
check_return:
	if (stream_write_pending())
		goto repeat;
	switch (mode) {
	case COMMAND_MODE:
		if (cp_numread <= 0 && cbo == 0) /* command complete */
			return 0;
		if (cbo && lines < bot.lines - 1) {
			int num = check_buf_for_items(command_buf, cbo,
				NULL, 0);
			if (num > 0) {
				num = MIN(num, bot.lines - 1 - lines);
				cbo = check_buf_for_items(command_buf, cbo,
					&add_output_line, num);
				lines += num;
			}
			if (lines >= bot.lines - 1)
				print_in_bar(COLOR_MSG,
					"Line %i-%i. Hit q to quit, any other "
					"key for next page", total_lines + 1,
					total_lines + lines);
		}
		ret = wgetch(top.win);
		if (ret != ERR && ret != KEY_RESIZE) {
			total_lines += lines;
			lines = 0;
		}
		if (ret == 'q') {
			cbo = 0;
			return -1;
		}
		break;
	case GETCH_MODE:
		ret = wgetch(top.win);
		if (ret != ERR && ret != KEY_RESIZE)
			return ret;
		break;
	case EXTERNAL_MODE:
		if (external_cmd_died) {
			external_cmd_died = 0;
			return 0;
		}
	}
	goto repeat;
}

/*
 * read from command pipe and print data to bot window
 */
static int send_output(pid_t pid)
{
	if (!command_pipe)
		return 0;
	if (do_select(COMMAND_MODE) >= 0)
		clog(INFO, "command complete\n");
	else
		clog(NOTICE, "command aborted\n");
	fclose(command_pipe);
	command_pipe = NULL;
	print_in_bar(COLOR_MSG, " ");
	return 1;
}

/*
 * exec client command and print output to status bar.
 */
static void print_executing_msg(char *cmd)
{
	outputf(COLOR_COMMAND, "%s\n", cmd);
	print_in_bar(COLOR_MSG, "executing client command, hit q to abort\n", 
		cmd);
	wattron(bot.win, COLOR_PAIR(COLOR_OUTPUT));
}

static int client_cmd(char *cmd,...)
{
	pid_t pid;
	print_executing_msg(cmd);
	command_pipe = popen_read_client_pid(&pid, &cmd);
	return send_output(pid);
}

static int client_cmd_cmdline(char *cmdline)
{
	pid_t pid;

	print_executing_msg(cmdline);
	command_pipe = popen_read_client_cmdline_pid(&pid, cmdline);
	return send_output(pid);
}

/*
 * exec command and print output to bot win
 */
static int display_cmd(char *cmd)
{
	pid_t pid;

	print_in_bar(COLOR_MSG, "executing display command, hit q to abort");
	outputf(COLOR_COMMAND, "%s\n", cmd);
	wattron(bot.win, COLOR_PAIR(COLOR_OUTPUT));

	command_pipe = popen_read_cmdline_pid(&pid, cmd);
	if (command_pipe)
		return send_output(pid);
	return -1;
}

/*
 * shutdown curses and stat pipe before executing external commands
 */
static int external_cmd(char *cmd)
{
	if (external_cmd_pid)
		return -1;
	shutdown_curses();
	exec_cmdline_pid_bg(&external_cmd_pid, cmd);
	do_select(EXTERNAL_MODE);
	init_curses();
	return 0;
}

static void com_ll_decr(int key)
{
	if (args_info.loglevel_arg <= DEBUG) {
		print_in_bar(COLOR_ERRMSG, 
			"loglevel already at maximal verbosity\n"); 
		return;
	}
	args_info.loglevel_arg--;
	print_in_bar(COLOR_MSG, "loglevel set to %d\n", args_info.loglevel_arg);
}

static void com_ll_incr(int key)
{
	if (args_info.loglevel_arg >= EMERG) {
		print_in_bar(COLOR_ERRMSG, 
			"loglevel already at miminal verbosity\n"); 
		return;
	}
	args_info.loglevel_arg++;
	print_in_bar(COLOR_MSG, "loglevel set to %d\n", args_info.loglevel_arg);
}

/*
 * reread configuration, terminate on errors
 */
static void com_reread_conf(int key)
{
	char *cf =configfile_exists();

	if (!cf) {
		clog(WARNING, "There is no configuration to read\n");
		return;
	}
	clog(DEBUG, "Rereading command line options\n");
	cmdline_parser(_argc, _argv, &args_info);
	clog(DEBUG, "Reading config file %s\n", cf);
	cmdline_parser_configfile(cf, &args_info, 1, 1, 0);
	clog(NOTICE, "configuration read\n");
	if (check_key_map_args() < 0)
		finish(EXIT_FAILURE);
}

static void com_help(int key)
{
	int i, j = 0;
	char keymaps[MAXLINE], *p = keymaps;

	for (i = 0; i < args_info.key_map_given; ++i) {
		char *s = args_info.key_map_arg[i], *handler, *arg,
			*desc = "", tmp[MAXLINE];
		
		strcpy(tmp, s);
		if (!split_key_map(tmp, &handler, &arg))
			return;
		switch(*handler) {
			case 'i':
				handler = "internal";
				desc = command_list[find_cmd_byname(arg)].description;
				break;
			case 'x': handler = "external"; break;
			case 'd': handler = "display "; break;
			case 'p': handler = "para    "; break;
		}
		outputf(COLOR_MSG, "%s\t%s\t%s%s\t%s\n", tmp, handler, arg,
			strlen(arg) < 8? "\t" : "",
			desc);
		if (i > 0 && !(i % (bot.lines - 1))) {
			print_in_bar(COLOR_MSG, "Hit q to abort, any other "
				"key for next page");
			if (do_select(GETCH_MODE) == 'q') {
				print_in_bar(COLOR_MSG, "Help aborted");
				return;
			}
		}
	}
	*p = '\0';
	j = i;
	for (i = 0; command_list[i].handler; i++) {
		struct gui_command gc = command_list[i];
		
		outputf(COLOR_MSG, "%s\tinternal\t%s\t%s%s\n", gc.key, gc.name,
			strlen(gc.name) < 8? "\t" : "",
			gc.description);
		j++;
		if (!(j % (bot.lines - 1)))
			if (do_select(GETCH_MODE) < 0)
				return;
	}
	print_in_bar(COLOR_MSG, "try :help or :help command for more info");
	return;
}

static void com_sl(int key)
{
	client_cmd("sl", "10", stat_items[STREAM].content, NULL);
}

static void com_toggle_auto_decode(int key)
{
	int ad = args_info.auto_decode_flag;

	ad = ad? 0 : 1;
		
	if (ad)
		clog(NOTICE, "activating auto decoding\n");
	else
		clog(NOTICE, "deactivating auto decoding\n");
	args_info.auto_decode_flag = ad;
}

static void com_shrink_top_win(int key)
{
	if (!top_lines_ok(top.lines - 1)) {
		clog(NOTICE, "Cannot decrease top window\n");
		return;
	}
	init_wins(top.lines - 1);
	wclear(top.win);
	print_all_items();
	print_in_bar(COLOR_MSG, "decreased top window");
}

static void com_enlarge_top_win(int key)
{
	if (!top_lines_ok(top.lines + 1)) {
		clog(NOTICE, "Cannot increase top window");
		return;
	}
	init_wins(top.lines + 1);
	wclear(top.win);
	print_all_items();
	print_in_bar(COLOR_MSG, "increased top window");
}

static void com_version(int key)
{
	print_in_bar(COLOR_MSG, "para_gui " VERSION " \"" CODENAME "\"");
}

static void com_query_stream(int key)
{
	client_cmd("strq", stat_items[STREAM].content, NULL);
}

static void com_quit(int key)
{
	finish(0);
}

static void com_refresh(int key)
{
	shutdown_curses();
	init_curses();
}

static void handle_command(int c)
{
	int i;

	/* first check user's key bindings */
	for (i = 0; i < args_info.key_map_given; ++i) {
		char tmp[MAXLINE], *handler, *arg;

		strcpy(tmp, args_info.key_map_arg[i]);
		if (!split_key_map(tmp, &handler, &arg))
			return;
		if (!strcmp(tmp, km_keyname(c))) {
			if (*handler == 'd') {
				display_cmd(arg);
				return;
			}
			if (*handler == 'x') {
				external_cmd(arg);
				return;
			}
			if (*handler == 'p') {
				client_cmd_cmdline(arg);
				return;
			}
			if (*handler == 'i') {
				int num = find_cmd_byname(arg);
				if (num >= 0)
					command_list[num].handler(c);
				return;
			}
		}
	}
	/* not found, check internal key bindings */
	for (i = 0; command_list[i].handler; i++) {
		if (!strcmp(km_keyname(c), command_list[i].key)) {
			command_list[i].handler(c);
			return;
		}
	}
	
	//outputf(COLOR_ERRMSG, "Unknown key: %i\n", c);
	print_in_bar(COLOR_ERRMSG, "key is not bound, press ? for help");
}

static int init_stream_io(void)
{
	int i, j, ret = -1;
	char *p, *q = NULL;

	i = 0;
	while (i < args_info.stream_write_cmd_given) {
		q = strdup(args_info.stream_write_cmd_arg[i]);
		p = strchr(q, ':');
		if (!p) {
			fprintf(stderr, "%s: invalid stream_write_cmd\n", 
				__func__);
			goto out;
		}
		*p = '\0';
		p++;
		for (j = 0; siol[j].name; j++) {
			if (strcmp(q, siol[j].name))
				continue;
			siol[j].wcmd = args_info.stream_write_cmd_arg[i] 
				+ (p - q);
			break;
		}
		free(q);
		q = NULL;
		i++;
	}
	//fprintf(stderr, "%s: invalid stream_read_cmd\n", __func__);
	i = 0;
	//fprintf(stderr, "%s: %d\n", __func__, args_info.stream_read_cmd_given);
	while (i < args_info.stream_read_cmd_given) {
		q = strdup(args_info.stream_read_cmd_arg[i]);
		p = strchr(q, ':');
		if (!p) {
			fprintf(stderr, "%s: invalid stream_read_cmd\n", 
				__func__);
			goto out;
		}
		*p = '\0';
		p++;
		for (j = 0; siol[j].name; j++) {
			if (strcmp(q, siol[j].name))
				continue;
			siol[j].rcmd = args_info.stream_write_cmd_arg[i] 
				+ (p - q);
			break;
		}
		free(q);
		q = NULL;
		i++;
	}

	for (j = 0; siol[j].name; j++) {
		siol[j].buf = malloc(siol[j].bufsize * sizeof(char));
		siol[j].buflen = 0;
	}
	ret = 1;
out:
	if (q)
		free(q);
	return ret;
}

int main(int argc, char *argv[])
{
	int ret;
	char *cf;

	_argc = argc;
	_argv = argv;

	if (cmdline_parser(argc, argv, &args_info)) {
		fprintf(stderr, "Parse error while reading command line\n");
		exit(EXIT_FAILURE);
	}
	if (check_key_map_args() < 0) {
		fprintf(stderr, "Invalid key map\n");
		exit(EXIT_FAILURE);
	}
	cf = configfile_exists();
	if (!cf && args_info.config_file_given) {
		fprintf(stderr, "Can't read config file %s\n",
			args_info.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (cf)
		cmdline_parser_configfile(cf, &args_info, 0, 0, 0);
	if (check_key_map_args() < 0) {
		fprintf(stderr, "Invalid key map in config file\n");
		exit(EXIT_FAILURE);
	}
	if (init_stream_io() < 0) {
		fprintf(stderr, "Invalid stream_read or stream_write command\n");
		exit(EXIT_FAILURE);
	}
	top.lines = args_info.top_lines_arg;
	signals_setup();
	initscr(); /* needed only once */
	init_curses();
	print_welcome();
	open_stat_pipe_msg();
	for (;;) {
		print_status_bar(NULL);
		ret = do_select(GETCH_MODE);
		if (!ret)
			continue;
		print_in_bar(COLOR_MSG, " ");
		handle_command(ret);
	}
}
